@@ -27,6 +27,7 @@ class OpenfeedClient(object):
 
         self.instrument_definitions = {}
         self.instruments_by_symbol = {}
+        self.subscription_symbol_by_marketid = {}
         self.snapshots = {}
 
         self.symbol_handlers = {}
@@ -37,6 +38,8 @@ class OpenfeedClient(object):
         self.on_connected = None
         self.on_disconnected = None
         self.on_error = None
+        self.on_login = None
+        self.on_logout = None
 
         websocket.enableTrace(self.debug)
 
@@ -62,28 +65,32 @@ class OpenfeedClient(object):
     def add_heartbeat_subscription(self, callback):
         """Subscribe to [Heartbeat] messages (keep alive)
 
-        [Heartbeat]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.HeartBeat
+        [Heartbeat]: https://docs.barchart.com/openfeed/#/proto?id=heartbeat
         """
         self.heartbeat_handlers.append(callback)
 
         return self
 
-    def add_symbol_subscription(self, symbol: Union[str, list], callback, service="REAL_TIME", subscription_type=["QUOTE"]):
+    def add_symbol_subscription(self, symbol: Union[str, list], callback, service="REAL_TIME", subscription_type=["QUOTE"], snapshot_interval_seconds=60):
         """Subscribe to [Market Data] by Barchart Symbols
 
-        Complete list of [SubscriptionTypes].
+        Complete list of [SubscriptionTypes]. List of [Service] types. List of [InstrumentTypes].
 
         Parameters
         ----------
         service: str, optional
-            Default is `REAL_TIME` for delayed market data, set it to `DELAYED`
+            Default is `REAL_TIME` for delayed market data, set it to `DELAYED`, or for snapshots set it as one of `REAL_TIME_SNAPSHOT` or `DELAYED_SNAPSHOT'
         callback: Callable
             Your callback function for Market Data messages
         subscription_type: list, optional
             Default is ['QUOTE']. Can contain any of: 'ALL', 'QUOTE', 'QUOTE_PARTICIPANT', 'DEPTH_PRICE', 'DEPTH_ORDER', 'TRADES', 'OHLC'
+        instrument_type: list, optional
+            Spreads and Options must be explicitly requested. Can contain any of: 'SPREAD', 'OPTION', 'FUTURE', 'FOREX', 'EQUITY', 'INDEX', 'MUTUAL_FUND', 'MONEY_MARKET', 'MONEY_MARKET_FUND'
 
-        [Market Data]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.MarketUpdate
-        [SubscriptionTypes]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.SubscriptionType
+        [Market Data]: https://docs.barchart.com/openfeed/#/proto?id=marketupdate
+        [SubscriptionTypes]: https://docs.barchart.com/openfeed/#/proto?id=subscriptiontype
+        [Service]: https://docs.barchart.com/openfeed/#/proto?id=service
+        [InstrumentTypes]: https://docs.barchart.com/openfeed/#/proto?id=instrumentdefinitioninstrumenttype
         """
         symbols = []
 
@@ -97,20 +104,24 @@ class OpenfeedClient(object):
                 self.symbol_handlers[sym] = []
 
             self.symbol_handlers[sym].append(
-                Listener(symbol=sym, callback=callback, service=service, subscription_type=subscription_type))
+                Listener(symbol=sym, callback=callback, service=service, subscription_type=subscription_type, snapshot_interval_seconds=snapshot_interval_seconds))
 
         if self.token is not None:
             self._send_message(
-                self.__create_subscription_request(symbols=symbols, service=service, subscription_type=subscription_type))
+                self.__create_subscription_request(symbols=symbols, service=service, subscription_type=subscription_type, snapshot_interval_seconds=snapshot_interval_seconds))
 
         return self
 
-    def add_exchange_subscription(self, exchange: Union[str, list], callback, service="REAL_TIME", subscription_type=["QUOTE"]):
-        """Subscribe to [Market Data] by Barchart Exchange code(s)
+    def add_exchange_subscription(self, exchange: Union[str, list], callback, service="REAL_TIME", subscription_type=["QUOTE"], instrument_type=[], snapshot_interval_seconds=60):
+        """Subscribe to [Market Data] by Barchart Exchange code(s).
+
+        Complete list of [SubscriptionTypes]. List of [Service] types.
 
         Note: your credentials must have the correct service level (FEED) for this operation.
 
-        [Market Data]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.MarketUpdate
+        [Market Data]: https://docs.barchart.com/openfeed/#/proto?id=marketupdate
+        [SubscriptionTypes]: https://docs.barchart.com/openfeed/#/proto?id=subscriptiontype
+        [Service]: https://docs.barchart.com/openfeed/#/proto?id=service
         """
         exchanges = []
 
@@ -124,18 +135,18 @@ class OpenfeedClient(object):
                 self.exchange_handlers[exch] = []
 
             self.exchange_handlers[exch].append(Listener(
-                exchange=exch, callback=callback, service=service, subscription_type=subscription_type))
+                exchange=exch, callback=callback, service=service, subscription_type=subscription_type, instrument_type=instrument_type, snapshot_interval_seconds=snapshot_interval_seconds))
 
         if self.token is not None:
             self._send_message(
-                self.__create_subscription_request(exchanges=exchanges, service=service, subscription_type=subscription_type))
+                self.__create_subscription_request(exchanges=exchanges, service=service, subscription_type=subscription_type, instrument_type=instrument_type, snapshot_interval_seconds=snapshot_interval_seconds))
 
         return self
 
     def request_available_exchanges(self, callback):
         """Request a list of available [Exchanges] for subscription.
 
-        [Exchanges]: https://github.com/openfeed-org/proto/blob/master/openfeed_api.proto#L159-L173
+        [Exchanges]: https://docs.barchart.com/openfeed/#/openfeed_streaming?id=requesting-exchanges
         """
         rid = random.getrandbits(32)
         req = self.__create_exchange_request(rid)
@@ -150,7 +161,7 @@ class OpenfeedClient(object):
     def request_instruments_for_exchange(self, exchange, callback):
         """Request a list of [Instrument Definitions] actively trading trading on an exchange.
 
-        [Instrument Definitions]: https://openfeed-org.github.io/documentation/Message%20Specification/#openfeed_instrumentproto
+        [Instrument Definitions]: https://docs.barchart.com/openfeed/#/proto?id=openfeed_instrumentproto
         """
 
         rid = random.getrandbits(32)
@@ -168,8 +179,8 @@ class OpenfeedClient(object):
 
         See [Instrument Request]
 
-        [Instrument Definitions]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.InstrumentDefinition
-        [Instrument Request]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.InstrumentRequest
+        [Instrument Definitions]: https://docs.barchart.com/openfeed/#/proto?id=openfeed_instrumentproto
+        [Instrument Request]: https://docs.barchart.com/openfeed/#/proto?id=instrumentrequest
         """
 
         rid = random.getrandbits(32)
@@ -186,22 +197,22 @@ class OpenfeedClient(object):
     def get_instrument_definitions(self):
         """Returns a dict of Openfeed [Instrument Definitions] keyed by MarketID
 
-        [Instrument Definitions]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.InstrumentDefinition
+        [Instrument Definitions]: https://docs.barchart.com/openfeed/#/proto?id=openfeed_instrumentproto
         """
         return self.instrument_definitions
 
     def get_instrument_definition(self, id):
         """Returns an [Instrument Definition] for a Market ID
 
-        [Instrument Definition]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.InstrumentDefinition
+        [Instrument Definition]: https://docs.barchart.com/openfeed/#/proto?id=openfeed_instrumentproto
         """
         return self.instrument_definitions[id].instrumentDefinition
 
     def get_instrument_definition_by_symbol(self, symbol):
         """Returns an [Instrument Definition] for a [Symbol] string
 
-        [Instrument Definition]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.InstrumentDefinition
-        [Symbol]: https://openfeed-org.github.io/documentation/Message%20Specification/#org.openfeed.InstrumentDefinition.Symbol
+        [Instrument Definition]: https://docs.barchart.com/openfeed/#/proto?id=openfeed_instrumentproto
+        [Symbol]: https://docs.barchart.com/openfeed/#/proto?id=instrumentdefinitionsymbol
         """
         return self.instruments_by_symbol[symbol].instrumentDefinition
 
@@ -211,7 +222,7 @@ class OpenfeedClient(object):
         self.ws.send(msg.SerializeToString(), websocket.ABNF.OPCODE_BINARY)
 
     def __reset(self):
-        self.ws.shutdown()
+        self.ws.close()
         self.token = None
         self.ws = websocket.WebSocket()
 
@@ -223,10 +234,22 @@ class OpenfeedClient(object):
                 raise Exception("Login has failed: ", msg)
 
             self.token = msg.loginResponse.token
+
+            self.__callback(self.on_login, msg)
             self.__send_existing_interest()
 
             return msg
 
+        def handleLogout(msg):
+
+            self.__callback(self.on_logout, msg)
+
+            if msg.logoutResponse.status.result == 115:
+                if self.debug:
+                    print("Logout due to concurrent logins:", msg)
+
+            return msg
+
         def handleHeartbeat(msg):
             self.__notify_heartbeat_listeners(msg)
             return msg
@@ -270,6 +293,7 @@ class OpenfeedClient(object):
                 raise Exception("Subscription has failed: ", msg)
 
             if len(msg.subscriptionResponse.symbol) > 0:
+                self.subscription_symbol_by_marketid[msg.subscriptionResponse.marketId] = msg.subscriptionResponse.symbol
                 self.__notify_symbol_listeners(
                     self.__create_instrument(msg.subscriptionResponse.symbol), msg)
             else:
@@ -284,8 +308,13 @@ class OpenfeedClient(object):
             for symbol in msg.instrumentDefinition.symbols:
                 self.instruments_by_symbol[symbol.symbol] = msg
 
+            for r in self.request_id_handlers.values():
+                if r.is_type("instrumentRequest"):
+                    if msg.instrumentDefinition.exchangeCode == r.request.instrumentRequest.exchange:
+                        r.callback(msg)
+
             self.__notify_exchange_listeners(
-                msg.instrumentDefinition.barchartExchangeCode, msg)
+                msg.instrumentDefinition.exchangeCode, msg)
             self.__notify_symbol_listeners(msg.instrumentDefinition, msg)
 
             return msg
@@ -293,7 +322,7 @@ class OpenfeedClient(object):
         def handleMarketUpdate(msg):
             inst = self.instrument_definitions[msg.marketUpdate.marketId].instrumentDefinition
 
-            self.__notify_exchange_listeners(inst.barchartExchangeCode, msg)
+            self.__notify_exchange_listeners(inst.exchangeCode, msg)
             self.__notify_symbol_listeners(inst, msg)
 
             return msg
@@ -303,7 +332,7 @@ class OpenfeedClient(object):
 
             self.snapshots[inst.marketId] = msg
 
-            self.__notify_exchange_listeners(inst.barchartExchangeCode, msg)
+            self.__notify_exchange_listeners(inst.exchangeCode, msg)
             self.__notify_symbol_listeners(inst, msg)
 
             return msg
@@ -311,13 +340,14 @@ class OpenfeedClient(object):
         def handleOHLC(msg):
             inst = self.instrument_definitions[msg.ohlc.marketId].instrumentDefinition
 
-            self.__notify_exchange_listeners(inst.barchartExchangeCode, msg)
+            self.__notify_exchange_listeners(inst.exchangeCode, msg)
             self.__notify_symbol_listeners(inst, msg)
 
             return msg
 
         handlers = {
             "loginResponse": handleLogin,
+            "logoutResponse": handleLogout,
             "heartBeat": handleHeartbeat,
             "exchangeResponse": handleExchangeRequest,
             "subscriptionResponse": handleSubscriptionResponse,
@@ -349,20 +379,21 @@ class OpenfeedClient(object):
 
         def on_error(ws, error):
             if self.debug:
-                print("WS Error: ", error)
+                print("WebSocket Error: ", error)
                 traceback.print_exc()
             self.__callback(self.on_error, error)
 
-        def on_close(ws):
+        def on_close(ws, close_status_code, close_msg):
             if self.debug:
-                print("WS Close")
+                print("WebSocket Close.", "Close Status Code:",
+                      close_status_code, "Close Message:", close_msg)
 
             self.__reset()
             self.__callback(self.on_disconnected, ws)
 
         def on_open(ws):
             if self.debug:
-                print("WS Open")
+                print("WebSocket Open")
 
             self._send_message(self.__create_login_request())
             self.__callback(self.on_connected, ws)
@@ -377,20 +408,23 @@ class OpenfeedClient(object):
 
     def __notify_symbol_listeners(self, instrument, msg):
 
-        # TODO review symbology handling, subbing by one and keying off the other can create unexpected results
-        # for example subscribing to "ZCYAIA40.CM" will come back with OF symbol (less the suffix) in `instrument.symbol`
-        # given the below, if the instrument contains duplicate `instrument.symbols`, the listeners will get duplicate callbacks
+        symbol_key = ""
+
+        if instrument.marketId in self.subscription_symbol_by_marketid:
+            symbol_key = self.subscription_symbol_by_marketid[instrument.marketId]
 
         for s in instrument.symbols:
             if s.symbol not in self.symbol_handlers or s.vendor != "Barchart":
-                return
+                continue
+            symbol_key = s.symbol
 
-            for cb in self.symbol_handlers[s.symbol]:
+        if symbol_key != "":
+            for cb in self.symbol_handlers[symbol_key]:
                 try:
                     cb.callback(msg)
                 except Exception as e:
                     if self.debug:
-                        print("Failed to notify `symbol` callback:", s, e)
+                        print("Failed to notify `symbol` callback:", symbol_key, e)
                     self.__callback(self.on_error, e)
 
     def __notify_exchange_listeners(self, exchange, msg):
@@ -428,7 +462,7 @@ class OpenfeedClient(object):
                 listeners_by_service = interest[l.service]
                 if l.key() not in listeners_by_service:
                     listeners_by_service[l.key()] = Listener(
-                        symbol=l.symbol, exchange=l.exchange, service=l.service, subscription_type=l.subscription_type)
+                        symbol=l.symbol, exchange=l.exchange, service=l.service, subscription_type=l.subscription_type, instrument_type=l.instrument_type, snapshot_interval_seconds=l.snapshot_interval_seconds)
                 else:
                     existing = listeners_by_service[l.key()]
                     existing.subscription_type = list(set(
@@ -438,13 +472,13 @@ class OpenfeedClient(object):
         for service in interest.keys():
             for i in interest[service].values():
                 self._send_message(
-                    self.__create_subscription_request(exchanges=i.exchanges(), symbols=i.symbols(), service=service, subscription_type=i.subscription_type))
+                    self.__create_subscription_request(exchanges=i.exchanges(), symbols=i.symbols(), service=service, subscription_type=i.subscription_type, instrument_type=i.get_instrument_types(), snapshot_interval_seconds=i.snapshot_interval_seconds))
 
         # send other rpc requests
         for req in self.request_id_handlers.values():
             req.send(self)
 
-    def __create_subscription_request(self, exchanges=[], symbols=[], service="REAL_TIME", subscription_type=["QUOTE"]):
+    def __create_subscription_request(self, exchanges=[], symbols=[], service="REAL_TIME", subscription_type=["QUOTE"], instrument_type=[], snapshot_interval_seconds=60):
         requests = []
 
         if len(exchanges) > 0:
@@ -453,7 +487,9 @@ class OpenfeedClient(object):
                     exchange=exch,
                     subscriptionType=[openfeed_api_pb2.SubscriptionType.Value(
                         t) for t in subscription_type],
-                    snapshotIntervalSeconds=60
+                    snapshotIntervalSeconds=snapshot_interval_seconds,
+                    instrumentType=[openfeed_instrument_pb2.InstrumentDefinition.InstrumentType.Value(
+                        t) for t in instrument_type]
                 ))
 
         if len(symbols) > 0:
@@ -461,7 +497,8 @@ class OpenfeedClient(object):
                 requests.append(openfeed_api_pb2.SubscriptionRequest.Request(
                     symbol=sym,
                     subscriptionType=[openfeed_api_pb2.SubscriptionType.Value(
-                        t) for t in subscription_type]
+                        t) for t in subscription_type],
+                    snapshotIntervalSeconds=snapshot_interval_seconds
                 ))
 
         of_req = openfeed_api_pb2.OpenfeedGatewayRequest(
@@ -533,12 +570,14 @@ class OpenfeedClient(object):
 
 
 class Listener(object):
-    def __init__(self, symbol="", exchange="", callback=None, service="REAL_TIME", subscription_type=["QUOTE"]):
+    def __init__(self, symbol="", exchange="", callback=None, service="REAL_TIME", subscription_type=["QUOTE"], instrument_type=[], snapshot_interval_seconds=60):
         self.symbol = symbol
         self.exchange = exchange
         self.callback = callback
         self.service = service
         self.subscription_type = subscription_type
+        self.instrument_type = instrument_type
+        self.snapshot_interval_seconds = snapshot_interval_seconds
 
     def key(self):
         if len(self.exchange) > 0:
@@ -555,6 +594,11 @@ class Listener(object):
             return [self.exchange]
         return []
 
+    def get_instrument_types(self):
+        if len(self.exchange) > 0:
+            return self.instrument_type
+        return []
+
     def has_same_interest(self, other):
         return collections.Counter(self.subscription_type) == collections.Counter(other.subscription_type)
 
@@ -579,6 +623,9 @@ class Request(object):
 
         of_client._send_message(self.request)
 
+    def is_type(self, request_type):
+        return request_type == self.request.WhichOneof("data")
+
 
 if __name__ == "__main__":
 
@@ -597,6 +644,8 @@ if __name__ == "__main__":
     of_client.on_error = lambda x: print("of-client: something went wrong:", x)
     of_client.on_disconnected = lambda x: print("of-client: disconnected")
     of_client.on_connected = lambda x: print("of-client: connected")
+    of_client.on_login = lambda x: print("of-client: login:", x)
+    of_client.on_logout = lambda x: print("of-client: logout:", x)
 
     # blocking mode
     of_client.start(blocking=False)
